@@ -1,7 +1,7 @@
 /* This is a utility program for listing SCSI devices and hosts (HBAs)
  * in the Linux operating system. It is applicable to kernel versions
  * 2.6.1 and greater.
- *  Copyright (C) 2003-2007 D. Gilbert
+ *  Copyright (C) 2003-2008 D. Gilbert
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
  *  the Free Software Foundation; either version 2, or (at your option)
@@ -26,7 +26,7 @@
 #include <linux/major.h>
 #include <time.h>
 
-static const char * version_str = "0.21  2008/07/10";
+static const char * version_str = "0.22  2008/12/26";
 
 #define NAME_LEN_MAX 260
 #define FT_OTHER 0
@@ -40,6 +40,9 @@ static const char * version_str = "0.21  2008/07/10";
 #define TRANSPORT_SAS_CLASS 4
 #define TRANSPORT_ISCSI 5
 #define TRANSPORT_SBP 6
+#define TRANSPORT_USB 7
+#define TRANSPORT_ATA 8         /* probably PATA, could be SATA */
+#define TRANSPORT_SATA 9        /* most likely SATA */
 
 static int transport_id = TRANSPORT_UNKNOWN;
 
@@ -51,7 +54,7 @@ static const char * sysfs_test_top = "/sys";
 static const char * proc_mounts = "/proc/mounts";
 static const char * bus_scsi_devs = "/bus/scsi/devices";
 static const char * class_scsi_dev = "/class/scsi_device/";
-static const char * scsi_host = "/class/scsi_host";
+static const char * scsi_host = "/class/scsi_host/";
 static const char * spi_host = "/class/spi_host/";
 static const char * spi_transport = "/class/spi_transport/";
 static const char * sas_host = "/class/sas_host/";
@@ -83,6 +86,7 @@ struct lsscsi_opt_coll {
         int kname;
         int transport;
         int verbose;
+        int protection;         /* data integrity */
 };
 
 
@@ -135,6 +139,7 @@ static struct option long_options[] = {
         {"long", 0, 0, 'l'},
         {"list", 0, 0, 'L'},
 /*      {"name", 0, 0, 'n'},    */
+        {"protection", 0, 0, 'p'},
 /*      {"sysfsroot", 1, 0, 'y'},       */
         {"transport", 0, 0, 't'},
         {"verbose", 0, 0, 'v'},
@@ -160,15 +165,17 @@ struct dev_node_list {
 };
 static struct dev_node_list* dev_node_listhead = NULL;
 
-struct sg_item_t {
+struct item_t {
         char name[NAME_LEN_MAX];
         int ft;
         int d_type;
 };
 
-static struct sg_item_t non_sg;
-static struct sg_item_t aa_sg;
-static struct sg_item_t aa_first;
+static struct item_t non_sg;
+static struct item_t aa_sg;
+static struct item_t aa_first;
+static struct item_t aa_sd;
+static struct item_t aa_block;
 
 static char sas_low_phy[NAME_LEN_MAX];
 static char sas_hold_end_device[NAME_LEN_MAX];
@@ -178,7 +185,47 @@ static const struct addr_hctl * iscsi_target_hct;
 static int iscsi_tsession_num;
 
 
+static void
+usage()
+{
+        fprintf(stderr, "Usage: lsscsi   [--classic] [--device] [--generic]"
+                        " [--help] [--hosts]\n"
+                        "\t\t[--kname] [--list] [--long] [--protection] "
+                        "[--transport]\n"
+                        "\t\t[--verbose] [--version] [<h:c:t:l>]\n");
+        fprintf(stderr, "  where:\n");
+        fprintf(stderr, "    --classic|-c      alternate output similar "
+                        "to 'cat /proc/scsi/scsi'\n");
+        fprintf(stderr, "    --device|-d       show device node's major + "
+                        "minor numbers\n");
+        fprintf(stderr, "    --generic|-g      show scsi generic device "
+                        "name\n");
+        fprintf(stderr, "    --help|-h         this usage information\n");
+        fprintf(stderr, "    --hosts|-H        lists scsi hosts rather than "
+                        "scsi devices\n");
+        fprintf(stderr, "    --kname|-k        show kernel name instead of "
+                        "device node name\n");
+        fprintf(stderr, "    --list|-L         additional information "
+                        "output one\n");
+        fprintf(stderr, "                      attribute=value per line\n");
+        fprintf(stderr, "    --long|-l         additional information "
+                        "output\n");
+        fprintf(stderr, "    --protection|-p   show data integrity "
+                        "(protection) information\n");
+        fprintf(stderr, "    --transport|-t    transport information for "
+                        "target or, if '--hosts'\n"
+                        "                      given, for initiator\n");
+        fprintf(stderr, "    --verbose|-v      output path names where data "
+                        "is found\n");
+        fprintf(stderr, "    --version|-V      output version string and "
+                        "exit\n");
+        fprintf(stderr, "    <h:c:t:l>         filter output list (def: "
+                        "'- - - -' (all))\n\n");
+        fprintf(stderr, "List SCSI devices or hosts, optionally with "
+                "additional information\n");
+}
 
+/* Compare <host:controller:target:lun> tuples (aka <h:c:t:l> or hctl) */
 static int
 cmp_hctl(const struct addr_hctl * le, const struct addr_hctl * ri)
 {
@@ -206,43 +253,9 @@ invalidate_hctl(struct addr_hctl * p)
         }
 }
 
-static void
-usage()
-{
-        fprintf(stderr, "Usage: lsscsi   [--classic] [--device] [--generic]"
-                        " [--help] [--hosts]\n"
-                        "\t\t[--kname] [--list] [--long] [--transport] "
-                        "[--verbose]\n"
-                        "\t\t[--version] [<h:c:t:l>]\n");
-        fprintf(stderr, "  where:\n");
-        fprintf(stderr, "    --classic|-c    alternate output similar "
-                        "to 'cat /proc/scsi/scsi'\n");
-        fprintf(stderr, "    --device|-d     show device node's major + "
-                        "minor numbers\n");
-        fprintf(stderr, "    --generic|-g    show scsi generic device "
-                        "name\n");
-        fprintf(stderr, "    --help|-h       this usage information\n");
-        fprintf(stderr, "    --hosts|-H      lists scsi hosts rather than "
-                        "scsi devices\n");
-        fprintf(stderr, "    --kname|-k      show kernel name instead of "
-                        "device node name\n");
-        fprintf(stderr, "    --list|-L       additional information "
-                        "output one\n");
-        fprintf(stderr, "                    attribute=value per line\n");
-        fprintf(stderr, "    --long|-l       additional information "
-                        "output\n");
-        fprintf(stderr, "    --transport|-t  transport information for "
-                        "target or, if '--hosts'\n"
-                        "                    given, for initiator\n");
-        fprintf(stderr, "    --verbose|-v    output path names where data "
-                        "is found\n");
-        fprintf(stderr, "    --version|-V    output version string and "
-                        "exit\n");
-        fprintf(stderr, "    <h:c:t:l>       filter output list (def: "
-                        "'- - - -' (all))\n\n");
-        fprintf(stderr, "List SCSI devices or hosts\n");
-}
-
+/* Return 1 for directory entry that is link or directory (other than
+ * a directory name starting with dot). Else return 0.
+ */
 static int
 first_scandir_select(const struct dirent * s)
 {
@@ -257,6 +270,80 @@ first_scandir_select(const struct dirent * s)
         return 1;
 }
 
+/* Return 1 for directory entry that is link or directory (other than a
+ * directory name starting with dot) that contains "block". Else return 0.
+ */
+static int
+block_scandir_select(const struct dirent * s)
+{
+        if ((DT_LNK != s->d_type) &&
+            ((DT_DIR != s->d_type) || ('.' == s->d_name[0])))
+                return 0;
+        if (strstr(s->d_name, "block")){
+                strncpy(aa_block.name, s->d_name, NAME_LEN_MAX);
+                aa_block.ft = FT_CHAR;  /* dummy */
+                aa_block.d_type =  s->d_type;
+        }
+        return 1;
+}
+
+/* scan for scsi_disk directory in  /sys/bus/scsi/devices/<h:c:i:l> */
+static int
+block_scan(const char * dir_name, const struct lsscsi_opt_coll * opts)
+{
+        char name[NAME_LEN_MAX];
+        struct dirent ** namelist;
+        int num, k;
+
+        num = scandir(dir_name, &namelist, block_scandir_select, NULL);
+        if (num < 0) {
+                if (opts->verbose > 0) {
+                        snprintf(name, NAME_LEN_MAX, "scandir: %s", dir_name);
+                        perror(name);
+                }
+                return -1;
+        }
+        for (k = 0; k < num; ++k)
+                free(namelist[k]);
+        free(namelist);
+        return num;
+}
+
+static int
+sd_scandir_select(const struct dirent * s)
+{
+        if ((DT_LNK != s->d_type) &&
+            ((DT_DIR != s->d_type) || ('.' == s->d_name[0])))
+                return 0;
+        if (strstr(s->d_name, "scsi_disk")){
+                strncpy(aa_sd.name, s->d_name, NAME_LEN_MAX);
+                aa_sd.ft = FT_CHAR;  /* dummy */
+                aa_sd.d_type =  s->d_type;
+        }
+        return 1;
+}
+
+static int
+sd_scan(const char * dir_name, const struct lsscsi_opt_coll * opts)
+{
+        char name[NAME_LEN_MAX];
+        struct dirent ** namelist;
+        int num, k;
+
+        num = scandir(dir_name, &namelist, sd_scandir_select, NULL);
+        if (num < 0) {
+                if (opts->verbose > 0) {
+                        snprintf(name, NAME_LEN_MAX, "scandir: %s", dir_name);
+                        perror(name);
+                }
+                return -1;
+        }
+        for (k = 0; k < num; ++k)
+                free(namelist[k]);
+        free(namelist);
+        return num;
+}
+
 /* scan for directory entry that is either a symlink or a directory */
 static int
 scan_for_first(const char * dir_name, const struct lsscsi_opt_coll * opts)
@@ -513,7 +600,8 @@ find_sysfsroot()
         return res;
 }
 
-/* Return 1 if directory, else 0 */
+/* If 'dir_name'/'base_name' is a directory chdir to it. If that is successful
+   return 1, else 0 */
 static int
 if_directory_chdir(const char * dir_name, const char * base_name)
 {
@@ -533,7 +621,10 @@ if_directory_chdir(const char * dir_name, const char * base_name)
         return 0;
 }
 
-/* Return 1 if directory, else 0 */
+/* If 'dir_name'/generic is a directory chdir to it. If that is successful
+   return 1. Otherwise look a directory of the form
+   'dir_name'/scsi_generic:sg<n> and if found chdir to it and return 1.
+   Otherwise return 0. */
 static int
 if_directory_ch2generic(const char * dir_name)
 {
@@ -565,7 +656,9 @@ if_directory_ch2generic(const char * dir_name)
         return 0;
 }
 
-/* Return 1 if found, else 0 if problems */
+/* If 'dir_name'/'base_name' is found places corresponding value in 'value'
+ * and returns 1 . Else returns 0.
+ */
 static int
 get_value(const char * dir_name, const char * base_name, char * value,
           int max_value_len)
@@ -738,8 +831,49 @@ exit:
         return match_found;
 }
 
+/* Fetch USB device name string (form "<b>-<p1>[.<p2>]+:<c>.<i>") given
+ * either a SCSI host name or devname (i.e. "h:c:t:l") string. If detected
+ * return 'b' (pointer to start of USB device name string which is null
+ * terminated), else return NULL.
+ */
+static char *
+get_usb_devname(const char * hname, const char * devname, char * b, int b_len)
+{
+        char buff[NAME_LEN_MAX];
+        char bf2[NAME_LEN_MAX];
+        int len;
+        const char * np;
+        char * cp;
+        char * c2p;
+
+        strcpy(buff, sysfsroot);
+        if (hname) {
+                strcat(buff, scsi_host);
+                np = hname;
+        } else if (devname) {
+                strcat(buff, class_scsi_dev);
+                np = devname;
+        } else
+                return NULL;
+        if (if_directory_chdir(buff, np) && getcwd(bf2, NAME_LEN_MAX) &&
+            strstr(bf2, "usb")) {
+                if (b_len > 0)
+                        b[0] = '\0';
+                if ((cp = strstr(bf2, "/host"))) {
+                        len = (cp - bf2) - 1;
+                        if ((len > 0) && ((c2p = memrchr(bf2, '/', len)))) {
+                                len = cp - ++c2p;
+                                snprintf(b, b_len, "%.*s", len, c2p);
+                        }
+                }
+                return b;
+        }
+        return NULL;
+}
+
 /*  Parse colon_list into host/channel/target/lun ("hctl") array,
- *  return 1 if successful, else 0 */
+ *  return 1 if successful, else 0.
+ */
 static int
 parse_colon_list(const char * colon_list, struct addr_hctl * outp)
 {
@@ -767,14 +901,16 @@ parse_colon_list(const char * colon_list, struct addr_hctl * outp)
         return 1;
 }
 
-/* Fetch initiator (port) wwn(s) or identifier if available. Return 1 if
-   successful, 0 otherwise */
+/* Check host associated with 'devname' for known transport types. If so set
+   transport_id, place a string in 'b' and return 1. Otherwise return 0. */
 static int
 transport_init(const char * devname, /* const struct lsscsi_opt_coll * opts, */
                int b_len, char * b)
 {
         char buff[NAME_LEN_MAX];
+        char wd[NAME_LEN_MAX];
         int off;
+        char * cp;
         struct stat a_stat;
 
         /* SPI host */
@@ -831,7 +967,6 @@ transport_init(const char * devname, /* const struct lsscsi_opt_coll * opts, */
         /* SAS class representation */
         strcpy(buff, sysfsroot);
         strcat(buff, scsi_host);
-        strcat(buff, "/");
         strcat(buff, devname);
         strcat(buff, "/device/sas/ha");
         if ((stat(buff, &a_stat) >= 0) && S_ISDIR(a_stat.st_mode)) {
@@ -852,7 +987,6 @@ transport_init(const char * devname, /* const struct lsscsi_opt_coll * opts, */
                 /* resolve SCSI host device */
                 strcpy(buff, sysfsroot);
                 strcat(buff, scsi_host);
-                strcat(buff, "/");
                 strcat(buff, devname);
                 strcat(buff, "/device");
                 if (readlink(buff, buff2, sizeof(buff2)) <= 0)
@@ -876,7 +1010,7 @@ transport_init(const char * devname, /* const struct lsscsi_opt_coll * opts, */
                 strcat(buff, buff2);
 
                 /* read the FireWire host's EUI-64 */
-                if (!get_value(buff, "host_id/guid", buff2, sizeof(buff)) ||
+                if (!get_value(buff, "host_id/guid", buff2, sizeof(buff2)) ||
                     strlen(buff2) != 18)
                         break;
                 snprintf(b, b_len, "sbp:%s", buff2 + 2);
@@ -895,9 +1029,39 @@ transport_init(const char * devname, /* const struct lsscsi_opt_coll * opts, */
 //           Hmmm, probably would like SAM-4 ",i,0x" notation here.
                 return 1;
         }
+        /* USB host? */
+        cp = get_usb_devname(devname, NULL, wd, sizeof(wd) - 1);
+        if (cp) {
+                transport_id = TRANSPORT_USB;
+                snprintf(b, b_len, "usb: %s", cp);
+                return 1;
+        }
+        /* ATA or SATA host, crude check: driver name */
+        strcpy(buff, sysfsroot);
+        strcat(buff, scsi_host);
+        strcat(buff, devname);
+        if (get_value(buff, "proc_name", wd, sizeof(wd))) {
+                if (0 == strcmp("ahci", wd)) {
+                        transport_id = TRANSPORT_SATA;
+                        snprintf(b, b_len, "sata:");
+                        return 1;
+                } else if (strstr(wd, "ata")) {
+                        if (0 == memcmp("sata", wd, 4)) {
+                                transport_id = TRANSPORT_SATA;
+                                snprintf(b, b_len, "sata:");
+                                return 1;
+                        }
+                        transport_id = TRANSPORT_ATA;
+                        snprintf(b, b_len, "ata:");
+                        return 1;
+                }
+        }
         return 0;
 }
 
+/* Given the transport_id of a SCSI host (initiator) associated with
+ * 'path_name' output additional information.
+ */
 static void
 transport_init_longer(const char * path_name,
                       const struct lsscsi_opt_coll * opts)
@@ -1026,11 +1190,23 @@ transport_init_longer(const char * path_name,
         case TRANSPORT_ISCSI:
                 printf("  transport=iSCSI\n");
 // >>>       This is the multi-line host output for iSCSI. Anymore to
-//           add here? [From /sys/class/scsi_host/hostN/device/iscsi_host:hostN directory]
+//           add here? [From
+//           /sys/class/scsi_host/hostN/device/iscsi_host:hostN directory]
                 break;
         case TRANSPORT_SBP:
                 printf("  transport=sbp\n");
                 break;
+        case TRANSPORT_USB:
+                printf("  transport=usb\n");
+                printf("  device_name=%s\n", get_usb_devname(cp, NULL,
+                       value, NAME_LEN_MAX));
+                break;
+        case TRANSPORT_ATA:
+                printf("  transport=ata\n");
+                break;
+        case TRANSPORT_SATA:
+                printf("  transport=sata\n");
+                break;
         default:
                 if (opts->verbose > 1)
                         fprintf(stderr, "No transport information\n");
@@ -1038,8 +1214,9 @@ transport_init_longer(const char * path_name,
         }
 }
 
-/* Fetch target port wwn(s) or identifier if available. Return 1 if
-   successful, 0 otherwise */
+/* Attempt to determine the transport type of the SCSI device (LU) associated
+   with 'devname'. If found set transport_id, place string in 'b' and return
+   1. Otherwise return 0. */
 static int
 transport_tport(const char * devname,
                 /* const struct lsscsi_opt_coll * opts, */ int b_len, char * b)
@@ -1055,6 +1232,7 @@ transport_tport(const char * devname,
 
         if (! parse_colon_list(devname, &hctl))
                 return 0;
+
         /* SAS host? */
         strcpy(buff, sysfsroot);
         strcat(buff, sas_host);
@@ -1173,9 +1351,39 @@ transport_tport(const char * devname,
 //           (UTF-8) excluding trailing null.
                 return 1;
         }
+        /* USB device? */
+        cp = get_usb_devname(NULL, devname, wd, sizeof(wd) - 1);
+        if (cp) {
+                transport_id = TRANSPORT_USB;
+                snprintf(b, b_len, "usb: %s", cp);
+                return 1;
+        }
+        /* ATA or SATA device, crude check: driver name */
+        strcpy(buff, sysfsroot);
+        strcat(buff, scsi_host);
+        len = strlen(buff);
+        snprintf(buff + len, NAME_LEN_MAX - len, "host%d", hctl.h);
+        if (get_value(buff, "proc_name", wd, sizeof(wd))) {
+                if (0 == strcmp("ahci", wd)) {
+                        transport_id = TRANSPORT_SATA;
+                        snprintf(b, b_len, "sata:");
+                        return 1;
+                } else if (strstr(wd, "ata")) {
+                        if (0 == memcmp("sata", wd, 4)) {
+                                transport_id = TRANSPORT_SATA;
+                                snprintf(b, b_len, "sata:");
+                                return 1;
+                        }
+                        transport_id = TRANSPORT_ATA;
+                        snprintf(b, b_len, "ata:");
+                        return 1;
+                }
+        }
         return 0;
 }
 
+/* Given the transport_id of the SCSI device (LU) associated with 'devname'
+   output additional information. */
 static void
 transport_tport_longer(const char * devname,
                        const struct lsscsi_opt_coll * opts)
@@ -1384,6 +1592,17 @@ transport_tport_longer(const char * devname,
                 if (opts->verbose > 2)
                         printf("fetched from directory: %s\n", buff);
                 break;
+        case TRANSPORT_USB:
+                printf("  transport=usb\n");
+                printf("  device_name=%s\n", get_usb_devname(NULL, devname,
+                       value, NAME_LEN_MAX));
+                break;
+        case TRANSPORT_ATA:
+                printf("  transport=ata\n");
+                break;
+        case TRANSPORT_SATA:
+                printf("  transport=sata\n");
+                break;
         default:
                 if (opts->verbose > 1)
                         fprintf(stderr, "No transport information\n");
@@ -1670,7 +1889,7 @@ one_sdev_entry(const char * dir_name, const char * devname,
                         else if (!get_dev_node(wd, dev_node, typ))
                                 snprintf(dev_node, NAME_MAX, "-       ");
 
-                        printf("%s", dev_node);
+                        printf("%-9s", dev_node);
                         if (opts->dev_maj_min) {
                                 if (get_value(wd, "dev", value, NAME_LEN_MAX))
                                         printf("[%s]", value);
@@ -1709,6 +1928,58 @@ one_sdev_entry(const char * dir_name, const char * devname,
                 else
                         printf("  -");
         }
+
+        if (opts->protection) {
+                int kernel_dif_support = 0;
+                if (sd_scan(buff,opts)) {
+                        if (if_directory_chdir(buff,aa_sd.name)) {
+                                char value[NAME_LEN_MAX];
+                                char sddir[NAME_LEN_MAX];
+                                strncpy(sddir,buff,NAME_LEN_MAX);
+                                strcat(sddir,"/");
+                                strcat(sddir,aa_sd.name);
+                                if (!get_value(sddir, "protection_type", value, 
+                                        NAME_LEN_MAX)) {
+                                        /* kernel < 2.6.27 */
+                                        if (opts->verbose)
+                                                printf(" No Data Integrity "
+                                                                "Support\n");
+                                } else {
+                                        kernel_dif_support = 1;
+                                        if (strncmp(value, "0", 1))
+                                                printf("  DIF/Type%1s ",value);
+                                        else
+                                                printf("  -         ");
+                                }
+                        } else {
+                                printf("  -         ");
+                        }
+                }
+
+                if (kernel_dif_support && block_scan(buff,opts)) {
+                        if (if_directory_chdir(buff,aa_block.name)) {
+                                char value[NAME_LEN_MAX];
+                                char blkdir[NAME_LEN_MAX];
+                                strncpy(blkdir,buff,NAME_LEN_MAX);
+                                strcat(blkdir,"/");
+                                strcat(blkdir,aa_block.name);
+                                if (if_directory_chdir(blkdir,"integrity")) {
+                                        if (!get_value(".", "format", value, 
+                                                                NAME_LEN_MAX)) {
+                                                if (opts->verbose)
+                                                        printf(" No Data "
+                                                                "Integrity "
+                                                                "Support\n");
+                                        } else {
+                                                printf(" %-17s",value);
+                                        }
+                                } else {
+                                        printf(" %-17s","-   ");
+                                }
+                        }
+                }
+        }
+
         printf("\n");
         if (opts->long_opt > 0)
                 longer_d_entry(buff, devname, opts);
@@ -1783,6 +2054,7 @@ sdev_scandir_sort(const void * a, const void * b)
         return cmp_hctl(&left_hctl, &right_hctl);
 }
 
+/* List SCSI devices (LUs). */
 static void
 list_sdevices(const struct lsscsi_opt_coll * opts)
 {
@@ -1818,6 +2090,7 @@ list_sdevices(const struct lsscsi_opt_coll * opts)
         free(namelist);
 }
 
+/* List host (initiator) attributes when --long given (one or more times). */
 static void
 longer_h_entry(const char * path_name, const struct lsscsi_opt_coll * opts)
 {
@@ -2071,7 +2344,7 @@ one_filter_arg(const char * arg, struct addr_hctl * filtp)
         return 0;
 }
 
-/* Return 0 if able to decode, otheriwse 1 */
+/* Return 0 if able to decode, otherwise 1 */
 static int
 decode_filter_arg(const char * a1p, const char * a2p, const char * a3p,
                   const char * a4p, struct addr_hctl * filtp)
@@ -2148,7 +2421,7 @@ main(int argc, char **argv)
         while (1) {
                 int option_index = 0;
 
-                c = getopt_long(argc, argv, "cdghHklLtvV", long_options,
+                c = getopt_long(argc, argv, "cdghHklLptvV", long_options,
                                 &option_index);
                 if (c == -1)
                         break;
@@ -2178,6 +2451,9 @@ main(int argc, char **argv)
                 case 'L':
                         opts.long_opt += 3;
                         break;
+                case 'p':
+                        ++opts.protection;
+                        break;
                 case 't':
                         ++opts.transport;
                         break;
